package guru.kake.xmlp

import java.io.FileInputStream
import javax.xml.namespace.QName
import javax.xml.stream.XMLEventReader
import javax.xml.stream.XMLInputFactory
import javax.xml.stream.events.XMLEvent

/** DSL marker */
@DslMarker
annotation class DSL

/**
 * Implements a DSL-language for event-based parsing of XML files.
 */
@DSL
class XMLParser private constructor(path: String) {
    companion object {
        /**
         * Parse an XML-file using an event-based interface
         * @param path path to file
         * @param init lambda to execute on each XML event
         */
        fun parse(path: String, init: XMLParser.() -> Unit): Unit = XMLParser(path).init(init)
    }

    private val factory: XMLInputFactory = XMLInputFactory.newInstance()
    private val reader: XMLEventReader = factory.createXMLEventReader(FileInputStream(path))
    private val elements: MutableMap<String, Element> = mutableMapOf()

    private fun init(init: XMLParser.() -> Unit) {
        @Suppress("UNUSED_EXPRESSION") // IntelliJ IDEA bug: https://youtrack.jetbrains.com/issue/KT-21282
        init()
        while (reader.hasNext()) {
            parse(reader.nextEvent())
        }
    }

    private fun parse(event: XMLEvent) {
        when {
            event.isStartElement -> {
                elements[event.asStartElement().name.localPart.toString()]?.parseStart(event)
            }
            event.isEndElement -> {
                elements[event.asEndElement().name.localPart.toString()]?.parseEnd()
            }
        }
    }

    /**
     * Specify code to execute based on element name
     * @param name Element name
     * @param f code to execute on elements matching name
     */
    fun element(name: String, f: Element.() -> Unit) {
        elements[name] = Element(reader).also { it.f() }
    }

    /**
     * XML element
     * @param reader [XMLEventReader] to be used for peeking at the next element in [content].
     */
    @DSL
    inner class Element(private val reader: XMLEventReader) {
        private var startFunc: (() -> Unit)? = null
        private var endFunc: (() -> Unit)? = null
        private var event: XMLEvent? = null

        internal fun parseStart(event: XMLEvent) {
            this.event = event
            startFunc?.invoke()
        }

        internal fun parseEnd() { endFunc?.invoke() }

        /**
         * Specify code to execute when encountering a starting tag of the specified element
         * @param f code to execute
         */
        fun start(f: () -> Unit) {
            startFunc = f
        }

        /**
         * Specify code to execute when encountering a closing tag of the specified element
         * @param f code to execute
         */
        fun end(f: () -> Unit) {
            endFunc = f
        }

        /**
         * Get the content of the current element
         * @return content of current tag or null
         */

        fun content(): String? {
            return try {
                reader.peek().asCharacters().data
            } catch (e: ClassCastException) {
                null
            }
        }

        /**
         * Get attribute of current element if it exists or null. Only works within a start element.
         * @param attribute attribute name
         * @return value of attribute or null
         */
        fun attribute(attribute: String): String? = event!!.asStartElement().getAttributeByName(QName(attribute))?.value
    }
}